#pragma once

#include <chrono>
#include <map>
#include <string>
#include "logging/Logger.hpp"

namespace precice {
namespace utils {

/// Represents an event that can be started and stopped.
/** Additionally to the duration there is a special property that can be set for a event.
A property is a a key-value pair with a numerical value that can be used to trace certain events,
like MPI calls in an event. It is intended to be set by the user. */
class Event
{
public:
  /// Default clock type. All other chrono types are derived from it.
  using Clock = std::chrono::steady_clock;

  /// Name used to identify the timer. Events of the same name are accumulated to
  std::string name;

  /// Allows to put a non-measured (i.e. with a given duration) Event to the measurements.
  Event(std::string eventName, Clock::duration eventDuration);

  /// Creates a new event and starts it, unless autostart = false, synchronize processes, when barrier == true
  /** Use barrier == true with caution, as it can lead to deadlocks. */
  Event(std::string eventName, bool barrier = false, bool autostart = true);

  /// Stops the event if it's running and report its times to the EventRegistry
  ~Event();

  /// Starts an event. If it's already started it has no effect.
  void start(bool barrier = false);

  /// Stops an event. If it's already stopped it has no effect.
  void stop(bool barrier = false);

  /// Gets the duration of the event.
  Clock::duration getDuration();

private:
  Clock::time_point starttime;
  Clock::time_point stoptime;
  Clock::duration duration = Clock::duration::zero();
  bool isStarted = false;
  bool _barrier = false;

  static logging::Logger _log;
};


/// Class that aggregates durations for a specific event.
class EventData
{
public:
  /// Adds an events data.
  void put(Event* event);

  /// Get the average duration of all events so far.
  int getAvg();

  /// Get the maximum duration of all events so far
  int getMax();

  /// Get the minimum duration of all events so far
  int getMin();

  /// Get the total duration of all events so far
  int getTotal();

  /// Get the number of all events so far
  int getCount();

  /// Get the time percentage that the total time of this event took in relation to the globalDuration.
  int getTimePercentage(Event::Clock::duration globalDuration);

private:
  int count = 0;
  Event::Clock::duration total = Event::Clock::duration::zero();
  Event::Clock::duration max   = Event::Clock::duration::min();
  Event::Clock::duration min   = Event::Clock::duration::max();

};


/// High level object that stores data of all events.
/** Call EventRegistry::intialize at the beginning of your application and
EventRegistry::finalize at the end. Event timings will be usuable without calling this
function at all, but global timings as well as percentages do not work this way.  */
class EventRegistry
{
public:
  /// Sets the global start time
  /**
   * @param[in] applicationName A name that is added to the logfile to distinguish different participants
   */
  static void initialize(std::string appName = "");

  /// Sets the global end time
  static void finalize();

  /// Clears the registry. needed for tests
  static void clear();

  /// Finalizes the timings and calls print. Can be used as a crash handler to still get some timing results.
  static void signal_handler(int signal);

  /// Records the event.
  static void put(Event* event);

  /// Adds a value to the global property store.
  /** An existing value is added. */ 
  static void addProp(std::string property, double value);

  /// Sets a value in the global property store
  /** An existing value is overwritten. */ 
  static void setProp(std::string property, double value);

  /// Prints a verbose report to stdout and a terse one to EventTimings-AppName.log
  static void printAll();

  /// Prints the result table to an arbitrary stream.
  /** terse enables a more machine readable format with one event per line, seperated by whitespace. */
  static void print(std::ostream &out, bool terse = false);

  /// Convenience function: Prints to std::cout
  static void print(bool terse = false);

  /// Convenience function: Prints to filename
  static void print(std::string filename, bool terse = false);

  static void printGlobalDuration();

private:
  static bool initialized;
  static Event::Clock::time_point globalStart;
  static Event::Clock::time_point globalStop;
  static std::map<std::string, EventData> events;

  /// Map of additional properties that can be set by the user.
  static std::map<std::string, double> properties;

  /// A name that is added to the logfile to distinguish different participants
  static std::string applicationName;
};

}} // namespace precice::utils
